"""Reasoner API models."""
import hashlib

from typing import Optional, Callable

from pydantic import constr, Field, parse_obj_as

from .base_model import BaseModel
from .utils import HashableSequence, HashableSet
from .results import Result, Results
from .qgraph import QueryGraph
from .kgraph import KnowledgeGraph
from .shared import LogEntry, LogLevel
from .workflow import Workflow
from .auxgraphs import AuxiliaryGraphs


class Message(BaseModel):
    """Message."""

    query_graph: Optional[QueryGraph] = Field(
        None,
        title="query graph",
        nullable=True,
    )
    knowledge_graph: Optional[KnowledgeGraph] = Field(
        None,
        title="knowledge graph",
        nullable=True,
    )
    results: Optional[Results] = Field(
        None,
        title="list of results",
        nullable=True,
    )
    auxiliary_graphs: Optional[AuxiliaryGraphs] = Field(
        None, title="dict of auxiliary graphs", nullable=True
    )

    class Config:
        title = "message"
        extra = "forbid"

    def parse_obj(obj):
        message = parse_obj_as(Message, obj)
        qgraph = None
        kgraph = None
        results = None
        auxgraphs = None
        if "query_graph" in obj.keys() and obj["query_graph"] is not None:
            qgraph = QueryGraph.parse_obj(obj["query_graph"])
        if "knowledge_graph" in obj.keys() and obj["knowledge_graph"] is not None:
            kgraph = KnowledgeGraph.parse_obj(obj["knowledge_graph"])
        if "results" in obj.keys() and obj["results"] is not None:
            results = Results.parse_obj(obj["results"])
        if "auxiliary_graphs" in obj.keys() and obj["auxiliary_graphs"] is not None:
            auxgraphs = AuxiliaryGraphs.parse_obj(obj["auxiliary_graphs"])
        m = Message(
            query_graph=qgraph,
            knowledge_graph=kgraph,
            results=results,
            auxiliary_graphs=auxgraphs,
        )
        if m.knowledge_graph:
            m._normalize_kg_edge_ids()
        m.update(message)
        return m

    def update(self, other: "Message"):
        if hash(self.query_graph) != hash(other.query_graph):
            raise NotImplementedError("Query graph merging not supported yet")
        # Make a copy because normalization will modify results
        other = other.copy(deep=True)

        if other.knowledge_graph:
            if not self.knowledge_graph:
                self.knowledge_graph = KnowledgeGraph(nodes=[], edges=[])
            # Normalize edges of incoming KG
            # This will place KG edge keys into the same hashing system
            # So that equivalence is determined by hash collision
            other._normalize_kg_edge_ids()
            # The knowledge graph can now be udated because edge keys will be
            # hashed using the same method. The knowledge graph update method
            # will handle concatenating properties when necessary.
            self.knowledge_graph.update(other.knowledge_graph)
        if other.results:
            if self.results:
                self.results.update(other.results)
            else:
                self.results = other.results
        if other.auxiliary_graphs:
            if self.auxiliary_graphs:
                self.auxiliary_graphs.update(other.auxiliary_graphs)
            else:
                self.auxiliary_graphs = other.auxiliary_graphs

    def _normalize_kg_edge_ids(self):
        """
        Replace edge IDs with a hash of the edge object
        """
        self._update_kg_edge_ids(
            lambda edge: hashlib.blake2b(
                hash(edge).to_bytes(16, byteorder="big", signed=True), digest_size=6
            ).hexdigest(),
        )

    def _update_kg_edge_ids(self, update_func: Callable):
        """
        Replace edge IDs using the specified function
        """

        # Mapping of old to new edge IDs
        edge_id_mapping = {}

        # Make a copy of the edge keys because we're about to change them
        for edge_id in list(self.knowledge_graph.edges.keys()):
            edge = self.knowledge_graph.edges.pop(edge_id)
            new_edge_id = update_func(edge)

            edge_id_mapping[edge_id] = new_edge_id

            # Update knowledge graph
            self.knowledge_graph.edges[new_edge_id] = edge

        # Update auxiliary graphs
        if self.auxiliary_graphs:
            for auxiliary_graph in self.auxiliary_graphs.values():
                if auxiliary_graph.edges:
                    for aux_edge in auxiliary_graph.edges:
                        auxiliary_graph.edges.discard(aux_edge)
                        auxiliary_graph.edges.add(
                            edge_id_mapping.get(aux_edge, aux_edge)
                        )

        # Update results
        if self.results:
            for result in self.results:
                if result and result.analyses:
                    for analysis in result.analyses:
                        for edge_binding_list in analysis.edge_bindings.values():
                            for eb in edge_binding_list:
                                eb.id = edge_id_mapping[eb.id]


class Query(BaseModel):
    """Request."""

    message: Message = Field(
        ...,
        title="message",
    )
    log_level: Optional[LogLevel] = Field(
        None,
        title="log_level",
        nullable=True,
    )
    workflow: Optional[Workflow]

    class Config:
        title = "query"
        extra = "allow"
        schema_extra = {"x-body-name": "request_body"}


class AsyncQuery(BaseModel):
    """AsyncQuery."""

    callback: constr(regex=r"^https?://") = Field(..., format="uri")
    message: Message = Field(
        ...,
        title="message",
    )
    log_level: Optional[LogLevel] = Field(
        None,
        title="log_level",
        nullable=True,
    )
    workflow: Optional[Workflow]

    class Config:
        title = "query"
        extra = "allow"
        schema_extra = {"x-body-name": "request_body"}


class Response(BaseModel):
    """Response."""

    message: Message = Field(
        ...,
        title="message",
    )

    logs: Optional[HashableSequence[LogEntry]] = Field(None, nullable=True)

    status: Optional[str] = Field(None, nullable=True)

    workflow: Optional[Workflow]

    class Config:
        title = "response"
        extra = "allow"


class AsyncQueryResponse(BaseModel):
    """ "Async Query Response."""

    status: Optional[str] = Field(None, nullable=True)

    description: Optional[str] = Field(None, nullable=True)

    job_id: str = Field(..., title="job id")

    class Config:
        title = "async query response"
        extra = "allow"


class AsyncQueryStatusResponse(BaseModel):
    """Async Query Status Response."""

    status: str = Field(..., title="status")

    description: str = Field(..., title="description")

    logs: Optional[HashableSet[LogEntry]] = Field(None, nullable=True)

    response_url: Optional[str] = Field(None, nullable=True)

    class Config:
        title = "async query status response"
        extra = "allow"



"""Auxiliary Graphs model"""
from typing import Optional

from pydantic import Field, parse_obj_as

from .base_model import BaseModel
from .utils import HashableMapping, HashableSet
from .shared import Attribute, CURIE


class AuxiliaryGraph(BaseModel):
    """Auxiliary Graph"""

    edges: HashableSet[str] = Field(..., title="edges in auxiliary graph")

    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "auxiliary graph"
        extra = "allow"


class AuxiliaryGraphs(BaseModel):
    """Auxiliary Graphs"""

    __root__: Optional[HashableMapping[str, AuxiliaryGraph]]

    class Config:
        title = "auxiliary graphs"
        extra = "allow"

    def update(self, other):
        self.__root__.update(other.__root__)

    def values(self):
        return self.__root__.values()

    def items(self):
        return self.__root__.items()

    def keys(self):
        return self.__root__.keys()

    def parse_obj(obj):
        auxiliary_graphs = parse_obj_as(AuxiliaryGraphs, obj)
        graphs = AuxiliaryGraphs()
        graphs.__root__ = HashableMapping[str, AuxiliaryGraph]()
        for id, graph in obj.items():
            graphs.__root__[id] = AuxiliaryGraph.parse_obj(graph)
        graphs.update(auxiliary_graphs)
        return graphs



"""Results models."""
from typing import Optional

from pydantic import Field, parse_obj_as

from .base_model import BaseModel
from .utils import HashableMapping, HashableSet, HashableSequence
from .shared import Attribute, CURIE


class EdgeBinding(BaseModel):
    """Edge binding."""

    id: str = Field(
        ...,
        title="knowledge graph id",
    )

    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "edge binding"
        schema_extra = {
            "example": {
                "id": "string",
            },
        }
        extra = "allow"


class Analysis(BaseModel):
    """Analysis."""

    resource_id: CURIE = Field(
        ...,
        title="resource infores",
    )

    edge_bindings: HashableMapping[str, HashableSet[EdgeBinding]] = Field(
        ...,
        title="list of edge bindings",
    )

    score: Optional[float] = Field(
        None,
        format="float",
    )

    support_graphs: Optional[HashableSet[str]] = Field(None, nullable=True)

    scoring_method: Optional[str] = Field(None, nullable=True)

    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "analysis"
        extra = "allow"

    def __hash__(self) -> int:
        return hash(
            (
                self.resource_id,
                self.edge_bindings,
                self.score,
                self.support_graphs,
                self.scoring_method,
            )
        )

    def update(self, other):
        for k in self.edge_bindings.keys():
            eb = other.edge_bindings.get(k)
            if eb:
                self.edge_bindings[k].update(eb)
        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes
        if other.support_graphs:
            if self.support_graphs:
                self.support_graphs.update(other.support_graphs)
            else:
                self.support_graphs = other.support_graphs


class NodeBinding(BaseModel):
    """Node binding."""

    id: CURIE = Field(
        ...,
        title="knowledge graph id",
    )

    query_id: Optional[CURIE] = Field(None, title="query graph id")

    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "node binding"
        schema_extra = {
            "example": {
                "id": "x:string",
            },
        }
        extra = "allow"


class Result(BaseModel):
    """Result."""

    node_bindings: HashableMapping[str, HashableSet[NodeBinding]] = Field(
        ...,
        title="list of node bindings",
    )

    analyses: Optional[HashableSet[Analysis]] = Field(
        None, title="list of anlysis blocks", nullable=True
    )

    class Config:
        title = "result"
        extra = "allow"

    def update(self, other):
        if other.analyses:
            if self.analyses:
                for analysis in other.analyses:
                    check = True
                    for ana in self.analyses:
                        if analysis == ana:
                            ana.update(analysis)
                            check = False
                    if check:
                        self.analyses.add(analysis)
            else:
                self.analyses = other.analyses

    def __hash__(self) -> int:
        return hash(self.node_bindings)

    def parse_obj(obj):
        result = parse_obj_as(Result, obj)
        nbindings = HashableMapping.parse_obj(obj["node_bindings"])
        analyses = HashableSet[Analysis]()
        if "analyses" in obj.keys():
            for analysis in obj["analyses"]:
                analyses.add(Analysis.parse_obj(analysis))
        r = Result(node_bindings=nbindings, analyses=analyses)
        result.update(r)
        return result

    def combine_analyses_by_resource_id(self):
        combine = HashableMapping[str, Analysis]()
        analyses = HashableSequence.parse_obj([analysis for analysis in self.analyses])
        for i, analysis in enumerate(analyses):
            if analysis.resource_id not in combine:
                combine[analysis.resource_id] = analysis
            for j, analysis_to_compare in enumerate(analyses[i + 1 :]):
                if (
                    analysis.resource_id == analysis_to_compare.resource_id
                    and analysis != analysis_to_compare
                ):
                    combine[analysis.resource_id].update(analysis_to_compare)

        combined_analyses = HashableSet[Analysis]()
        for analysis in combine.values():
            combined_analyses.add(analysis)

        self.analyses = combined_analyses


class Results(BaseModel):
    """Results."""

    __root__: Optional[HashableSet[Result]]

    class Config:
        title = "results"
        extra = "allow"

    def add(self, result):
        results = self.__root__
        if result in results:
            # this is slow for larger results
            for original_result in results:
                if result == original_result:
                    original_result.update(result)
                    return
        else:
            results.add(result)

    def __len__(self):
        return len(self.__root__)

    def __iter__(self):
        return self.__root__.__iter__()

    def update(self, other):
        for result in other.__root__:
            self.add(result)

    def parse_obj(obj):
        results = parse_obj_as(Results, obj)
        for result in obj:
            results.add(Result.parse_obj(result))
        return results



"""Knowledge graph models."""
from typing import Optional

from pydantic import Field

from .shared import (
    Attribute,
    BiolinkEntity,
    BiolinkPredicate,
    CURIE,
    EdgeIdentifier,
    Qualifier,
    ResourceRoleEnum,
)
from .base_model import BaseModel
from .utils import HashableMapping, HashableSet


class Node(BaseModel):
    """Knowledge graph node."""

    categories: Optional[HashableSet[BiolinkEntity]] = Field(
        None,
        title="categories",
        nullable=True,
    )
    name: Optional[str] = Field(None, nullable=True)
    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "knowledge-graph node"
        schema_extra = {
            "example": {
                "category": "string",
            },
        }
        extra = "forbid"
        allow_population_by_field_name = True

    def update(self, other):
        if other.name:
            self.name = other.name

        if other.categories:
            if self.categories:
                self.categories.update(other.categories)
            else:
                self.categories = other.categories

        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes


class RetrievalSource(BaseModel):
    """A component of source retrieval provenance"""

    resource_id: CURIE = Field(..., title="infores for source")

    resource_role: ResourceRoleEnum = Field(..., title="source type")

    upstream_resource_ids: Optional[HashableSet[CURIE]] = Field(None, nullable=True)

    source_record_urls: Optional[HashableSet[str]] = Field(None, nullable=True)

    def __hash__(self) -> int:
        return hash((self.resource_id, self.resource_role))

    def update(self, other):
        if other.upstream_resource_ids:
            if self.upstream_resource_ids:
                self.upstream_resource_ids.update(other.upstream_resource_ids)
            else:
                self.upstream_resource_ids = other.upstream_resource_ids


class Edge(BaseModel):
    """Knowledge graph edge."""

    subject: CURIE = Field(
        ...,
        title="subject node id",
    )
    object: CURIE = Field(
        ...,
        title="object node id",
    )
    predicate: BiolinkPredicate = Field(..., title="edge predicate")
    sources: HashableSet[RetrievalSource] = Field(
        ..., title="list of source retrievals"
    )
    qualifiers: Optional[HashableSet[Qualifier]] = Field(None, nullable=True)
    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "knowledge-graph edge"
        extra = "forbid"

    def update(self, other):
        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes
        if other.sources:
            if self.sources:
                self.sources.update(other.sources)
            else:
                self.sources = other.sources

    def get_primary_knowedge_source(self):
        for source in self.sources:
            if source.resource_role == "primary_knowledge_source":
                return source.resource_id

    def __hash__(self) -> int:
        primary_knowledge_source = self.get_primary_knowedge_source()
        return hash(
            (
                self.subject,
                self.object,
                self.predicate,
                self.qualifiers,
                primary_knowledge_source,
            )
        )


class KnowledgeGraph(BaseModel):
    """Knowledge graph."""

    nodes: HashableMapping[CURIE, Node] = Field(
        ...,
        title="nodes",
    )
    edges: HashableMapping[EdgeIdentifier, Edge] = Field(
        ...,
        title="edges",
    )

    class Config:
        title = "knowledge graph"
        extra = "allow"

    def update(self, other):
        for key, value in other.nodes.items():
            existing = self.nodes.get(key, None)
            if existing:
                existing.update(value)
            else:
                self.nodes[key] = value

        for key, value in other.edges.items():
            existing = self.edges.get(key, None)
            if existing:
                existing.update(value)
            else:
                self.edges[key] = value



"""Shared models."""
from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
import string
from typing import Any, Optional

from pydantic import Field
from pydantic.types import ConstrainedStr

from .base_model import BaseModel
from .utils import HashableSequence


# TODO: potential add validation for structure of CURIE
class CURIE(str):
    """Compact URI."""


class ResourceRoleEnum(str, Enum):
    """Types of resources"""

    aggregator_knowledge_source = "aggregator_knowledge_source"
    primary_knowledge_source = "primary_knowledge_source"
    supporting_data_source = "supporting_data_source"


class KnowledgeType(str, Enum):
    "Knowledge Type."

    lookup = "lookup"
    inferred = "inferred"


class EdgeIdentifier(str):
    """Identifier for an edge in a knowledge graph"""


class RecursiveAttribute(BaseModel):
    """Attribute subattribute."""

    attribute_type_id: CURIE = Field(..., title="type")
    value: Any = Field(..., title="value")
    value_type_id: Optional[CURIE] = Field(
        None,
        title="value_type_id",
        nullable=True,
    )
    original_attribute_name: Optional[str] = Field(None, nullable=True)
    value_url: Optional[str] = Field(None, nullable=True)
    attribute_source: Optional[str] = Field(None, nullable=True)
    description: Optional[str] = Field(None, nullable=True)
    attributes: Optional[HashableSequence[RecursiveAttribute]] = Field(
        None, nullable=True
    )

    class Config:
        extra = "forbid"


class SubAttribute(BaseModel):
    """Attribute subattribute."""

    attribute_type_id: CURIE = Field(..., title="type")
    value: Any = Field(..., title="value")
    value_type_id: Optional[CURIE] = Field(
        None,
        title="value_type_id",
        nullable=True,
    )
    original_attribute_name: Optional[str] = Field(None, nullable=True)
    value_url: Optional[str] = Field(None, nullable=True)
    attribute_source: Optional[str] = Field(None, nullable=True)
    description: Optional[str] = Field(None, nullable=True)
    attributes: Optional[HashableSequence[RecursiveAttribute]] = Field(
        None, nullable=True
    )

    class Config:
        extra = "forbid"


class Attribute(BaseModel):
    """Node/edge attribute."""

    attribute_type_id: CURIE = Field(..., title="type")
    value: Any = Field(..., title="value")
    value_type_id: Optional[CURIE] = Field(
        None,
        title="value_type_id",
        nullable=True,
    )
    original_attribute_name: Optional[str] = Field(None, nullable=True)
    value_url: Optional[str] = Field(None, nullable=True)
    attribute_source: Optional[str] = Field(None, nullable=True)
    description: Optional[str] = Field(None, nullable=True)
    attributes: Optional[HashableSequence[SubAttribute]] = Field(None, nullable=True)

    class Config:
        extra = "forbid"


class BiolinkQualifier(ConstrainedStr):
    """Biolink Qualifier."""

    regex = re.compile("^biolink:[a-z][a-z_]*$")

    class Config:
        title = "biolink entity"


class Qualifier(BaseModel):
    """Edge qualifier."""

    qualifier_type_id: BiolinkQualifier = Field(..., title="type")

    qualifier_value: str = Field(..., title="value")


class BiolinkEntity(ConstrainedStr):
    """Biolink entity."""

    regex = re.compile("^biolink:[A-Z][a-zA-Z]*$")

    class Config:
        title = "biolink entity"


class BiolinkPredicate(ConstrainedStr):
    """Biolink predicate."""

    regex = re.compile("^biolink:[a-z][a-z_]*$")

    class Config:
        title = "biolink predicate"


class LogLevelEnum(str, Enum):
    """Log level."""

    error = "ERROR"
    warning = "WARNING"
    info = "INFO"
    debug = "DEBUG"


class LogLevel(BaseModel):
    """Log level."""

    __root__: LogLevelEnum


class LogEntry(BaseModel):
    """Log entry."""

    timestamp: Optional[datetime] = Field(None, nullable=True)
    level: Optional[LogLevel] = Field(None, nullable=True)
    code: Optional[str] = Field(None, nullable=True)
    message: Optional[str] = Field(None, nullable=True)

    class Config:
        extra = "allow"


RecursiveAttribute.update_forward_refs()



"""Reasoner-pydantic module."""

from .kgraph import KnowledgeGraph, Node, Edge
from .qgraph import (
    QueryGraph,
    QNode,
    QEdge,
    AttributeConstraint,
)
from .results import Result, NodeBinding, EdgeBinding, Results, Analysis
from .auxgraphs import AuxiliaryGraphs, AuxiliaryGraph
from .message import (
    Message,
    Query,
    Response,
    AsyncQuery,
    AsyncQueryResponse,
    AsyncQueryStatusResponse,
)
from .workflow import (
    Operation,
    Workflow,
)
from .shared import (
    Attribute,
    BiolinkEntity,
    BiolinkPredicate,
    BiolinkQualifier,
    CURIE,
    LogEntry,
    LogLevel,
)
from .metakg import (
    MetaEdge,
    MetaNode,
    MetaKnowledgeGraph,
    MetaAttribute,
)

components = [
    Attribute,
    BiolinkEntity,
    BiolinkPredicate,
    BiolinkQualifier,
    CURIE,
    Edge,
    EdgeBinding,
    KnowledgeGraph,
    LogEntry,
    Message,
    Node,
    NodeBinding,
    QEdge,
    QNode,
    Query,
    QueryGraph,
    AsyncQuery,
    Result,
    Response,
    AsyncQueryResponse,
    AsyncQueryStatusResponse,
    LogLevel,
    AttributeConstraint,
    MetaEdge,
    MetaNode,
    MetaKnowledgeGraph,
    MetaAttribute,
    Results,
    AuxiliaryGraph,
    AuxiliaryGraphs,
    Operation,
    Workflow,
    Analysis,
]



import json

from pydantic import BaseModel as PydanticBaseModel

from pydantic.class_validators import root_validator
from .utils import make_hashable


class BaseModel(PydanticBaseModel):
    """
    Custom base model for all classes

    This provides hash and equality methods.
    """

    def __hash__(self) -> int:
        """Hash function based on Pydantic implementation"""
        return hash((self.__class__, tuple(self.__dict__.values())))

    def __eq__(self, other) -> bool:
        """Equality function that calls hash function"""
        return self.__hash__() == other.__hash__()

    def update(self, other):
        """Update fields on this object with fields from other object"""
        raise NotImplementedError(
            f"Model {self.__class__.__name__} has no update method"
        )

    # After running validation on all known properties, make sure everything else is hashable
    @root_validator(allow_reuse=True, pre=False)
    def make_hashable_root(cls, values):
        # The root validator must take and return a dict
        return {k: make_hashable(v) for k, v in values.items()}

    def to_dict(self) -> dict:
        return json.loads(self.json())
